"""Method for sharing global sensor and port configuration values between
modules.
"""
import serial
portdefaults = {
    "port_baud": 9600,
    "port_stopbits": serial.STOPBITS_ONE,
    "port_parity": serial.PARITY_NONE,
    "port_timeout":  0.01,
    "virtual": False,
    "delimiter": r"\s",
    "encoding": "UTF-8"
    }
"""dict: default parameters passed to the :class:`serial.Serial` constructor for
communicating with the serial port.
"""

def _get_parser(config):
    """Returns a :class:`~configparser.ConfigParser` instance for the specified
    file.

    Args:
        config (ConfigParser): instance from which to extract the sensor list
          and port information. `str` is also allowed, in which case it
          should be the path to the config file to load.
    """
    try:
        from configparser import ConfigParser
    except ImportError: # pragma: no cover
        #Renaming of modules to lower case in python 3.
        from ConfigParser import ConfigParser

    if isinstance(config, str):
        parser = ConfigParser()
        parser.readfp(open(config))    
    else: # pragma: no cover
        parser = config               

    return parser
        
_sensors = {}
"""dict: keys are sensor names, values are :class:`liveserial.config.Sensor`
instances.
"""
_sensors_parsed = False
"""bool: when True, we have already scanned the config file for sensor
settings.
"""
def _load_sensors(config):
    """Loads all the sensors from the specified config file.

    Args:
        config (ConfigParser): instance from which to extract the sensor list.
          `str` is also allowed, in which case it
          should be the path to the config file to load.
    """
    global _sensors, _sensors_parsed
    if not _sensors_parsed:
        parser = _get_parser(config)
        if parser is not None:
            #Now that we have the thread, we can add configuration for each of the
            #sensors in the config file.
            from fnmatch import fnmatch
            for section in parser.sections():
                if fnmatch(section, "sensor.*"):
                    name = section[len("sensor."):]
                    _sensors[name] = Sensor(None,name,**dict(parser.items(section)))
                
        _sensors_parsed = True
    
def sensors(config, sensor=None, port=None, monitor=None):
    """Returns the list of :class:`liveserial.config.Sensor` instances that have
    the specified port.

    Args:
        config (ConfigParser): instance from which to extract the sensor list
          and port information. `str` is also allowed, in which case it
          should be the path to the config file to load.
        port (str): name of the port to configure for.
    """
    _load_sensors(config)
    global _sensors
    if port is not None:
        result = {}
        for sensor_n, instance in _sensors.items():
            if instance.port == port:
                instance.monitor = monitor
                if monitor is not None:
                    instance.port = monitor.port
                result[sensor_n] = instance
    elif sensor is not None:
        if sensor in _sensors:
            result = _sensors[sensor]
        else:
            result = None

    return result
                    
_ports = {}
"""dict: keys are port names, values are updated parameter dictionaries that can
be passed to the :class:`liveserial.monitor.ComMonitorThread` constructor.
"""
_ports_parsed = False
"""bool: when True, we have already examined the config file for port settings.
"""
def _load_ports(config):
    """Loads the configured port information from the specified configuraton.
    """
    global _ports, _ports_parsed
    if not _ports_parsed:
        parser = _get_parser(config)
        if parser is None: # pragma: no cover
            _ports_parsed = True
            return
        
        from fnmatch import fnmatch
        for section in parser.sections():
            if fnmatch(section, "port.*"):
                name = section[len("port."):]
            else: # pragma: no cover
                continue

            params = portdefaults.copy()
            for option, value in params.items():
                #Override the value using the config value unless it doesn't
                #exist.
                if parser.has_option(section, option):
                    params[option] = parser.get(section, option)

            #Python's bool is interesting because bool('0') => True. So, we test
            #explicitly here for the option value the user set.
            import re
            if not isinstance(params["virtual"], bool):
                if re.match(r"\b\d+\b", params["virtual"]):
                    params["virtual"] = bool(int(params["virtual"]))
                elif re.match(r"[a-z]", params["virtual"][0], re.I):
                    params["virtual"] = params["virtual"][0].lower() == 't'

            _ports[name] = params
            
        _ports_parsed = True

def ports(config, port):
    """Returns the port configuration dictionary for the specified port name.

    Args:
        config (ConfigParser): instance from which to extract the port
          information. `str` is also allowed, in which case it
          should be the path to the config file to load.
        port (str): name of the port to return configuration for.
    """
    _load_ports(config)
    if port in _ports:
        return _ports[port]
    else: # pragma: no cover
        return portdefaults.copy()

_script = {}
"""dict: keys are command-line arguments usually accepted by the script when it
is run. Values are configured option values from the config file.
"""
_script_parsed = False
"""bool: when True, the script options have been parsed already.
"""
def script(config):
    """Returns the config options configured globally for the script.

    Args:
        config (ConfigParser): instance from which to extract the port
          information. `str` is also allowed, in which case it
          should be the path to the config file to load.
    """
    global _script, _script_parsed
    if not _script_parsed:
        parser = _get_parser(config)
        if parser is not None:
            for section in parser.sections():
                if section == "global":
                    _script = dict(parser.items("global"))

            #We also need to handle the types, since all the options are just
            #strings by default.
            from liveserial.livemon import script_options
            for name in _script:
                optname = "-{}".format(name)
                if optname in script_options and "type" in script_options[optname]:
                    caster = script_options[optname]["type"]
                    _script[name] = caster(_script[name])
        _script_parsed = True
                
    return _script
    
def _config_split(value, delim, cast=None):
    """Splits the specified value using `delim` and optionally casting the
    resulting items.

    Args:
        value (str): config option to split.
        delim (str): string to split the option value on.
        cast (function): to apply to each item after the split operation.
    """
    if value is None:
        return
    
    if delim is None: # pragma: no cover
        vals = value.split()
    else:
        vals = value.split(delim)
        
    if cast is not None:
        return list(map(cast, vals))
    else:
        return vals

_plot = {}
"""dict: of plot options; keys are ['line', 'axes', 'figure', 'label', 'ticks'];
values are dicts of matplotlib option values.

"""
_plot_parsed = False
"""bool: when True, we have parsed plot options already.
"""

def plot(config, element):
    """Returns the matplotlib configuration options for the specified plotting
    element.

    Args:
        config (ConfigParser): instance from which to extract the port
          information. `str` is also allowed, in which case it
          should be the path to the config file to load.
        element (str): one of ['line', 'axes', 'figure', 'label', 'ticks'];
          specifies which part of the plot the options will apply to.
    """
    global _plot, _plot_parsed
    if not _plot_parsed:
        parser = _get_parser(config)
        if parser is not None:
            from fnmatch import fnmatch
            for section in parser.sections():
                if fnmatch(section, "plot.*"):
                    name = section[len("plot."):]
                else: # pragma: no cover
                    continue

                _plot[name] = dict(parser.items(section))
        _plot_parsed = True

    if element in _plot:
        return _plot[element]
    else:
        return {}

def reset_config():
    """Resets the global config variables so that a session can be continued
    with a new config file.
    """
    global _sensors, _ports, _plot, _script
    global _sensors_parsed, _ports_parsed, _plot_parsed, _script_parsed
    _sensors = {}
    _sensors_parsed = False
    _ports = {}
    _ports_parsed = False
    _plot = {}
    _plot_parsed = False
    _script = {}
    _script_parsed = False

def _parse_transform(function):
    """Parses the transform function's fqdn to return the function that can
    actually transform the data.
    """
    if "numpy" in function: # pragma: no cover
        import numpy as np

    return eval(function)            

class Sensor(object):
    """Represents the configuration of a sensor on the serial port.

    Args:
        monitor (ComMonitorThread): parent instance that this sensor is being logged
          with.
        name (str): name of the sensor in the configuration file.
        key (str): sensor key as it will be written to the serial stream, or `None`
          if there isn't a key in the stream (i.e., only values).
        value_index (list): column index/indices of the value that will
          be plotted.
        dtype (list): of `str` or `type`; items must belong to ['key', int, float,
          str]. Represents the order in which values are found in a single line of
          data read from the serial port. Thus `W 129229 0.928379` would be given by
          ["key", int, float].
        label (str): for plots, what to put on the y-axis. Defaults to `name`.
        port (str): name of the port to read this sensor from. Defaults to
          :data:`ComMonitorThread.port`.
        logging (str): comma-separated list of columns indices (zero-based) to
            include in the log file. If not specified, then the default is to
            include *all* data columns in the log file.
        columns (str): comma-separated list of columns headings for the CSV file;
            these are written in the first row of the file. If excluded, they
            default to `Time` and `Value1`, `Value2`, etc.
        legends (str): if the comma-separated list in `value_index` includes more
            than one index, multiple lines are plotted on the same subplot. In that
            case, `legends` allows a comma-separated list of legend labels to be
            provided for each of those lines.
        sensors (str): comma-separated list of sensor names to include in the data
            vector that will be passed to `function` to be aggregated to a single
            value. This only applies to the case of aggregate sensors.
        function (str): name of a function to use to transform the data. Only
            applies to the case of aggregate sensors.
        kwargs (dict): additional keyword arguments supported that do not require
            special processing (i.e., are just simple string values).

    Attributes:
        options (dict): additional keyword arguments (or configurable options) for
          the sensor.
    """
    def __init__(self, monitor, name, key=None, value_index=None,
                 dtype=["key", "int", "float"], label=None, port=None,
                 logging=None, columns=None, legends=None, function=None,
                 sensors=None, **kwargs):

        self.monitor = monitor
        self.name = name
        self.key = key

        #We analyze the string values set for the dtypes to return the python
        #`types` that can cast strings to actual type values.
        self.dtype = []
        self._keyloc = None
        
        from six import string_types
        if isinstance(dtype, string_types):
            dtype = dtype.split(',')

        for i, sentry in enumerate(dtype):
            if sentry == "key":
                self._keyloc = i
                if key is None and port != "aggregate":
                    #For aggregate ports, we relax this condition since the
                    #sensors being aggregated have the dtypes specified
                    #correctly.
                    raise ValueError("You must specify a sensor key if 'key' "
                                     "is in the 'dtype' option list.")
            else:
                caster = eval(sentry)
                self.dtype.append(caster)

        self.value_index = _config_split(value_index, ',', int)
        self.label = name if label is None else label
        self.port = monitor.port if port is None else port
        self.logging = _config_split(logging, ',', int)
        self.columns = _config_split(columns, ',')
        self.legends = _config_split(legends, ',')
        self.sensors = _config_split(sensors, ',')
        
        if function is not None:
            self.transform = _parse_transform(function)
        else:
            self.transform = None
        
        self.options = kwargs
               
    def _cast(self, raw):
        """Casts all the values in the given list to their relevant data
        types. Assumes that the list has the correct format.
        
        Args:
            vals (list): string values from the split line to cast.
        """
        if (len(raw) != len(self.dtype)
            + (1 if self._keyloc is not None else 0)): # pragma: no cover
            return
        
        try:
            vals = []
            for iv, v in enumerate(raw):
                if iv != self._keyloc:
                    vals.append(self.dtype[len(vals)](v))

            #Previously, we were changing the order of the columns based on the
            #value index to make it easier for the plotter. Since we allow
            #multiple values to be plotted on the same subplot now, it is easier
            #to just not mangle them in the first place.
            return vals
        except ValueError: # pragma: no cover
            return None
        
    def parse(self, raw):
        """Parses a single line read from the serial port and returns a tuple of
        values.

        Args:
            raw (list): of split ASCII-encoded strings from the serial port. 
        
        Returns:
            list: of values parsed using :attr:`Sensor.dtype` casting.
            None: if the `key` was not found in the correct location.
        """
        result = None
        if self._keyloc is not None:
            if raw[self._keyloc] == self.key:
                result = self._cast(raw)
        elif self.key is None:
            result = self._cast(raw)
            
        return result



"""Methods for grabbing data and logging it to CSV files.
"""
from liveserial import msg
class Logger(object):
    """Logs data points from the serial port to CSV. The arguments to this class
    constructor are also available as attributes on the class instance.

    Args:
        interval (int): how often (in milliseconds) to read data from the serial
          buffer.
        dataqs (list): of :class:`multiprocessing.Queue`; stores the data read
          in from the serial port.
        errorqs (list): of :class:`multiprocessing.Queue`; stores any error
          raised during serial port reading.
        livefeed (monitor.LiveDataFeed): feed for storing the latest data points
          obtained from the serial port.
        method (str): one of ['average', 'last'] specifies how to aggregate multiple
          data points in the buffer.
        logdir (str): directory to place log files in for the sensors. If `None`,
          then data will *not* be logged to disk.
        logfreq (int): how often (in seconds) to write the accumulated data points
          to CSV.
        plotting (bool): when False, the values being read off will be printed if
          logging is not enabled.
        config (ConfigParser): global sensor configuration parser; if the
          argument is a `str`, then the file path.
        aggregate (bool): when True, the logger should check sensor config for
          aggregate ports to be handled; otherwise, aggregate ports are ignored.
    Attributes:
        timer (threading.Timer): executes calls to the serial port reader to get the
          latest data and push it to the live feed.    
        lastsave (float): timestamp indicating the last time the CSV file was
          appended to
        csvdata (dict): lists of sensor time and value readings, keyed by sensor
          identifier.
        config (ConfigParser or str): global sensor configuration parser. If
          `str`, then the config is loaded from the specified file path.
        aggregate (dict): keys are *aggregate* sensor names; values are functions
          that accept a dict of the latest physical sensor values, and return
          single, aggregated values for plotting.
    """
    def __init__(self, interval, dataqs, livefeed,
                 method="last", logdir=None, logfreq=10,
                 plotting=False, config=None, aggregate=False):
        
        self.interval = interval
        #Our first business is to make sure that we have only a list of *unique*
        #data queues to be querying.
        self.dataqs = []
        for dq in dataqs:
            if dq not in self.dataqs:
                self.dataqs.append(dq)
                
        self.livefeed = livefeed
        self.method = method

        from os import path, makedirs
        if logdir is not None:
            self.logdir = path.abspath(path.expanduser(logdir))
            if not path.isdir(self.logdir):
                msg.info("Making new directory path '{}'.".format(self.logdir))
                makedirs(self.logdir)
        else:
            self.logdir = None
        
        self.logfreq = logfreq
        self.lastsave = None
        self.csvdata = {}
        self.plotting = plotting
        self.timer = None
        self._timer_calls = 0
        """Number of times that the timer has executed during the application
        run.
        """
        self._cancel = False
        """When true, the main thread is trying to shut us down; don't start the
        timer again once it fires.
        """

        self.config = config
        if aggregate:
            self.aggregate = self._sensor_aggregates()
        else:
            self.aggregate = None

    def _sensor_aggregates(self):
        """Returns the transformations for each aggregate sensor so that they
        can be quickly compiled each time the logger timer executes.
        """
        from liveserial.config import sensors
        aggsensors = sensors(self.config, port="aggregate")
        result = {}
        for sensor, instance in aggsensors.items():
            if (instance.transform is not None and
                instance.sensors is not None):
                def aggfun(aggdata):
                    #We don't take the first index in the data element because
                    #it is the time stamp, which must be common to all of them.
                    v = []
                    t = None
                    for s in instance.sensors:
                        if s in aggdata:
                            v.append(aggdata[s][1:])
                            if t is None:
                                t = [aggdata[s][0]]

                    for vals in zip(*v):
                        t.append(instance.transform(vals))
                    return t
                        
                result[sensor] = aggfun

        return result
        
    def sensor_option(self, sensor, option, default=None, cast=None):
        """Returns the specified sensor option if available.

        Args:
            sensor (str): name of the sensor to return `option` for.
            option (str): option name.
            default: if the option is not configured, the default value to return.
            cast (function): if the raw value needs to be cast or transformed, the
                function to perform that transformation.
        """
        from liveserial.config import sensors, Sensor
        s = sensors(self.config, sensor)
        if isinstance(s, Sensor):
            if hasattr(s, option):
                value = getattr(s, option)
            elif option in s.options: # pragma: no cover
                value = s.options[option]
                
            if cast is not None: # pragma: no cover
                return cast(value)
            else:
                return value
        else: # pragma: no cover
            return default
                        
    def ready(self, delay=None):
        """Returns True once we have accumulated a few timer calls of data. This
        ensures that we know how many sensors are running on the same COM port.

        Args:
            delay (float): fraction of a second to wait before checking if the
              data is there.
        """
        if delay is not None:
            from time import sleep
            sleep(delay)
            
        havepts = False
        if len(self.csvdata) > 0:
            datapoints = sum([len(v) for v in self.csvdata.values()])/len(self.csvdata)
            havepts = datapoints > 2

        return self._timer_calls > 5 and (self.logdir is None or havepts)
        
    def start(self):
        """Starts a new timer for the configured interval to gather data from
        the serial stream.
        """
        if not self._cancel:
            from threading import Timer
            self.timer = Timer(self.interval, self._read_serial)
            self.timer.start()

    def stop(self):
        """Stops the automatic collection and logging of data. Cleans up threads.
        """
        self._cancel = True
        if self.timer is not None:
            self.timer.cancel()
        #Write whatever data is left over to CSV file.
        self._csv_append()
        
    def _read_serial(self):
        """Reads the latest buffered serial information and places it onto the live
        feed for the application.
        """
        self._timer_calls += 1
        from liveserial.monitor import get_all_from_queue
        sensedata = {}
        havedata = False
        #Just iterate over the various queues we have and process their data.
        for dataq in self.dataqs:
            for qdata in get_all_from_queue(dataq):
                sensor = qdata[0]              
                if sensor not in sensedata:
                    sensedata[sensor] = []
                sensedata[sensor].append(qdata[1:])
                havedata = True

        # We average/discard the data in the queue to produce the single entry that
        # will be posted to the livefeed.
        if havedata:
            if self.aggregate:
                aggdata = {}
                
            for sensor, qdata in sensedata.items():
                data = None
                if self.method == "average":
                    #We use the last data point's time stamp as the authoritative
                    #one for the averaged set.
                    tstamp = qdata[-1][0]
                    #For the values, we take a simple mean.
                    from numpy import mean
                    ldata = mean(qdata, axis=0)
                elif self.method == "last":
                    ldata = qdata[-1]
                    
                if ldata is not None:
                    self.livefeed.add_data(sensor, ldata)
                    
                if self.logdir is not None:
                    if sensor not in self.csvdata:
                        self.csvdata[sensor] = []
                    self.csvdata[sensor].extend(qdata)
                elif not self.plotting: # pragma: no cover
                    print("{}: {}".format(sensor, ldata))

                if self.aggregate:
                    aggdata[sensor] = ldata

            #Now that we have processed all the physical sensors, let's see if
            #we have any aggregate sensors that need to be processed. 
            if self.aggregate:
                for aggsense, aggfun in self.aggregate.items():
                    adata = aggfun(aggdata)
                    self.livefeed.add_data(aggsense, adata)
                    if self.logdir is not None:
                        if aggsense not in self.csvdata:
                            self.csvdata[aggsense] = []
                        self.csvdata[aggsense].append(adata)
                    elif not self.plotting: # pragma: no cover
                        print("{}: {}".format(aggsense, adata))
                        
            #Before we restart the timer again, see if we need to save the data to
            #CSV.
            self.save()
        self.start()

    def _csv_append(self):
        """Appends the new data points to the relevant CSV files for each of the
        sensor's whose data is being tracked.
        """
        from os import path
        import csv
        for sensor in self.csvdata:
            if self.logdir is None: # pragma: no cover
                #This should never fire because of checks further up the chain.
                #it is here as as sanity check to keep the file system clean.
                continue

            #Let's figure out how many columns to post in the header.
            if len(self.csvdata[sensor]) > 0:
                #We subtract 1 because the time is handled separately.
                N = len(self.csvdata[sensor][0])-1
            else: # pragma: no cover
                #No sense in writing to the file at this time; we have no
                #data to write!
                continue

            #Check if we have configuration settings available for this sensor.
            columns = None
            if self.config is not None:
                logids = self.sensor_option(sensor, "logging")
                    
                #We have one issue with the column labeling. We have switched
                #the important value to be in position 1, whereas it could be
                #anywhere in the list. Fix the order of the columns.
                cols = self.sensor_option(sensor, "columns")
                if logids is None and cols is not None: # pragma: no cover
                    #It is possible that the user will specify one or the other,
                    #but I want to limit the number of concurrent streams for
                    #the unit tests (especially for multi-port testing).
                    logids = list(range(len(cols)))

                vindex = self.sensor_option(sensor, "value_index", [0])

                #The other issue is that if the user limits the columns being
                #logged to include only a few columns, they will supply only a
                #few column labels.
                if (logids is not None and cols is not None and
                    len(logids) == len(cols)):
                    columns = {l: c for l, c in zip(logids, cols)}
                elif cols is not None: # pragma: no cover
                    msg.warn("A different number of column headings than "
                             "logging ids was specified in configuration.")
            else:
                logids = None
                vindex = [1]
                        
            if logids is None:
                #The user didn't say what to log, so we log everything.
                logids = list(range(1, N+1))

            logpath = path.join(self.logdir, "{}.csv".format(sensor))
            from os import linesep
            if not path.isfile(logpath):
                with open(logpath, 'wb') as f:
                    if columns is None:
                        columns = {li: "Value {}".format(i+1)
                                   for i, li in enumerate(logids)}
                    #Now, write the columns in the order specified in the
                    #logindex. However, we must remember that until now, we have
                    #kept the value in position 1, so we have to unmix that.
                    strcols = [columns[li] for li in logids]
                    header = "{}{}".format(','.join(["Time"] + strcols), linesep)
                    f.write(header.encode("ASCII"))

            from six import PY2
            mode = 'ab' if PY2 else 'a'
            kwds = {} if PY2 else {"newline": ''}
            with open(logpath, mode, **kwds) as f:
                #We log the full data stream from the sensor unless the logging is
                #limited by the configuration file.
                logids.insert(0, 0)
                writer = csv.writer(f)                    
                for idata in self.csvdata[sensor]:
                    if idata is not None:
                        #Sometimes, the aggregate data functions return None
                        #because one of the sensors didn't have data ready when
                        #the aggregation was performed. Just ignore those.
                        writer.writerow([idata[li] for li in logids])
                    
                #Since we appended the most recent data points, just reset the
                #lists of points to be empty again.
                self.csvdata[sensor] = []        
        
    def save(self):
        """Saves the logger's buffered points to a CSV file. If the file exists,
        then the data points are appended.
        """
        #We need to see if enough time has passed since the last
        from datetime import datetime
        from time import time
        if self.lastsave is not None:
            elapsed = (datetime.fromtimestamp(time()) -
                       datetime.fromtimestamp(self.lastsave)).total_seconds()
        else:
            elapsed = self.logfreq + 1

        if elapsed > self.logfreq:
            self._csv_append()
            self.lastsave = time()



"""Methods for plotting the real-time data feed.
"""
from liveserial.base import testmode
from matplotlib import cm
import matplotlib.animation as animation
import matplotlib

#We have to fiddle with the backends for Windows and Unix-based, otherwise we
#get unhandled exceptions or program-stopped working problems.
import os
if os.name != "nt":
    matplotlib.use("Agg" if testmode else "TkAgg")

from liveserial import msg
import numpy as np
from threading import Timer

def colorspace(size, cmap=cm.winter):
    """Returns an cycler over a linear color space with 'size' entries.
    
    :arg int size: the number of colors to define in the space.
    :returns: iterable cycler with 'size' colors.
    :rtype: itertools.cycle
    """
    from itertools import cycle
    import numpy as np
    rbcolors = cmap(np.linspace(0, 1, size))
    return (rbcolors, cycle(rbcolors))

class Plotter(animation.TimedAnimation):
    """Plots the live stream for each of the sensors in a subplot.

    Args:
        livefeed (monitor.LiveDataFeed): data feed with the latest data points to
          plot. Also an attribute on the class instance.
        interval (int): how often (in milliseconds) to redraw the plot with the
          latest plot values.
        maxlen (int): maximum number of time points kept for each subplot.    
        window (float): width of the plot window for sensors.
        testmode (bool): when True, the animator is not initialized so that a
          backend isn't required to run the unit tests.
        logger (log.Logger): logger instance for interacting with config
          parameters.

    Attributes:
        lines (dict): of :class:`matplotlib.lines.Line2D` being animated with
          the serial data; keyed by the sensor identifiers.
        axes (dict): of :class:`matplotlib.axes.Axes` being animated with the
          serial data; keyed by the sensor identifiers.
        ts (dict): of lists of the last `maxlen` sensor timestamp readings.
        ys (dict): of lists of the last `maxlen` sensor value readings. Keys for
          this dict are sensor names if only one value is specified in the
          sensor's `value_index` config option; otherwise, the keys are `(sensor,
          vindex)` tuple, where `vindex` is the zero-based, integer column index
          being plotted.

    """
    def __init__(self, livefeed, interval, maxlen=100, window=20,
                 testmode=False, logger=None):
        self.livefeed = livefeed
        self.interval = interval
        self.maxlen = maxlen
        self.window = window
        self.testmode = testmode
        self.logger = logger
        self.lines = {}
        self.axes = {}
        self.ts = {}
        self.ys = {}
        self._vindices = {}
        """Keys are sensor names; values are lists of value_index options from
        the configuration of each sensor.
        """
        self._plotorder = []
        """Sensor keys, ordered alphabetically; this is the order in which the
        subplots show up in the figure.
        """
        self._timer = None
        """Timer for unit testing the plotting code data acquisition.
        """
        
        #Find out how many subplots we will need; sort their keys for plotting.
        self._plotorder = sorted(self.livefeed.cur_data.keys(),
                                 key=lambda k: str(k))
        self._vindices = {s: self.logger.sensor_option(s, "value_index", [1])
                          for s in self._plotorder}

        if len(self._plotorder) == 0: # pragma: no cover
            raise ValueError("Live feed has no sensor data keys. "
                             "Can't setup plotting. Try fiddling with the "
                             "`-wait` parameter.")

        #For the plot styling, we use the config files. Logger has access to the
        #config file setting, so we just use that.
        from liveserial.config import plot
        import matplotlib.pyplot as plt
        from matplotlib.lines import Line2D

        #We are going to use a common time axis
        figopts = plot(self.logger.config, "figure")
        axesopts = plot(self.logger.config, "axes")
        if "figsize" not in figopts:
            figopts["figsize"] = (12, 3*len(self.livefeed.cur_data))
        else:
            figopts["figsize"] = tuple(map(float,figopts["figsize"].split(',')))
            
        fig, axes = plt.subplots(len(self.livefeed.cur_data), 1, sharex=True,
                                 squeeze=False, subplot_kw=axesopts, **figopts)

        from itertools import cycle
        from collections import deque
        totlines = sum([len(v) for v in self._vindices.values()])
        cspace = cycle(colorspace(totlines)[0])

        lineopts = plot(self.logger.config, "line")
        labelopts = plot(self.logger.config, "label")
        from six import string_types
        for isense, sensor in enumerate(self._plotorder):
            axes[isense,0].set_xlabel('t', **labelopts)
            if isinstance(sensor, string_types):
                label = self.logger.sensor_option(sensor, "label", sensor)
                port = self.logger.sensor_option(sensor, "port")
                if port is not None:
                    if port == "aggregate":
                        ylabel = "{} (agg.)".format(label)
                    else:
                        ylabel = "{} ({})".format(label, port)
                else: # pragma: no cover
                    ylabel = label
                axes[isense,0].set_ylabel(ylabel, **labelopts)
            else:
                axes[isense,0].set_ylabel("Auto {}".format(isense + 1),
                                          **labelopts) 

            legends = None
            if len(self._vindices[sensor]) > 1:
                legends = self.logger.sensor_option(sensor, "legends")

            for vi, vindex in enumerate(self._vindices[sensor]):
                if legends is not None:
                    legend = legends[vi]
                else:
                    legend = None
                    
                line = Line2D([], [], color=next(cspace), linewidth=2,
                              label=legend, **lineopts)
                axes[isense,0].add_line(line)
                axes[isense,0].set_xlim((0, window + 2.5))
                self.lines[(sensor, vindex)] = line
                self.ys[(sensor, vindex)] = deque(maxlen=self.maxlen)
                
            self.ts[sensor] = deque(maxlen=self.maxlen)
            self.axes[sensor] = axes[isense, 0]
            if legends is not None:
                self.axes[sensor].legend()

        tickopts = plot(self.logger.config, "ticks")
        if len(tickopts) > 0:
            plt.tick_params(**tickopts)
                
        from os import name
        if not self.testmode: # pragma: no cover
            if name != "nt":
                #Mac doesn't support blitting yet with its backends, so we have to
                #do the costly redraw at every iteration.
                animation.TimedAnimation.__init__(self, fig, blit=False)
            else:
                #If blitting is enabled, the plot labels don't update with
                #auto-scaling. Since our machines are fast enough, we make it
                #true by default. The user can override it using an option in
                #the config file if they need the extra speed.
                animopts = plot(self.logger.config, "animation")
                if "blit" not in animopts:
                    blit = True
                else:
                    blit = animopts["blit"] == 1
                    
                animation.TimedAnimation.__init__(self, fig, blit=blit)
        else:
            self._init_draw()
            self.new_frame_seq()
            self._timer = Timer(self.interval, self._draw_frame, (0,))
            self._timer.start()
            
        msg.info("Plotting animation configured.", 2)

    def new_frame_seq(self):
        return iter(range(self.maxlen))
        
    def _draw_frame(self, iframe):
        """Draws the latest frame for each sensor on the relevant plot.
        """
        for sensor in self._plotorder:
            #First, we get the latest data point from the live feed.
            ldata = self.livefeed.read_data(sensor)
            if len(ldata) < 2: # pragma: no cover
                #We don't have anything reasonable to plot; exit gracefully.
                continue
            
            t = ldata[0]
            self.ts[sensor].append(t)
            
            for vindex in self._vindices[sensor]:
                self.ys[(sensor, vindex)].append(ldata[vindex])
                ts, ys = self.ts[sensor], self.ys[(sensor, vindex)]
                self.lines[(sensor, vindex)].set_data(ts, ys)
            if t > self.window: # pragma: no cover
                # We don't want the tests to run long enough for this window to
                # kick in (at least for the moment).
                self.axes[sensor].set_xlim((self.ts[sensor][0], t + 2.5))

            self.axes[sensor].relim() # reset intern limits of the current axes 
            self.axes[sensor].autoscale_view()   # reset axes limits 
            
        self._drawn_artists = self.lines.values()
        if self.testmode:
            self._timer = Timer(self.interval, self._draw_frame, (0,))
            self._timer.start()
        
    def _init_draw(self):
        """Initializes all the subplot line objects to be empty."""
        for l in self.lines.values():
            l.set_data([], [])

    def stop(self):
        """Cleans up the timer when the plotter is running in test mode.
        """
        if self._timer:
            self._timer.cancel()



"""Module that sets up a virtual serial port for unit testing.
"""
from liveserial import msg
def examples():
    """Prints examples of using the script to the console using colored output.
    """
    script = "LIVE-SERIAL Virtual Serial Port Writer"
    explain = ("For testing the plotting and logging locally, it is useful to "
               "have virtual serial ports started on the local machine so "
               "that code can be tested without the real devices. This "
               "script creates threads that write to virtual ports.")
    contents = [(("Write 2 sensors to '/dev/tty.lscom-w' with formats `int`,"
                  "`float` and `int`,`int."), 
                 ("simport.py lscom-w -sensors K W -dtype K int float "
                  "W int int"),
                 ("Use '-seed' to change the random seed for the generated "
                  "values. ")),
                (("Write 2 sensor values to '/dev/tty.lscom-w' and "
                  "'/dev/tty.lscom-mw' simultaneously. Use default `int',"
                  "`float` data types for all but sensor `P`."),
                 ("simport.py lscom-w lscom-mw -sensors lscom-w K W "
                  "lscom-mw P S -dtype P float float"),
                 ("`None` can be used as a sensor name, in which case "
                  "a sensor without a key will be written to the stream."))]
    required = ("REQUIRED: virtual serial ports with the given names.")
    output = ("")
    details = ("For unix-based OS, the name is automatically changed to "
               "'/dev/tty.{}', where '{}' is the given name.")
    outputfmt = ("")

    msg.example(script, explain, contents, required, output, outputfmt, details)
    
def _parser_options():
    """Parses the options and arguments from the command line."""
    import argparse
    from liveserial import base
    parser = argparse.ArgumentParser(parents=[base.bparser],
                                     description="Real-time serial port plotter/logger.")
    parser.add_argument("ports", nargs="+",
                        help="Name of the port(s) to write random data to.")
    parser.add_argument("-sensors", nargs="+",
                        help=("Specify the port names and sensor names to "
                              "generate random data for. Format: "
                              "`-sensor portname sensor_0 sensor_1 None`"))
    parser.add_argument("-dtype", nargs="+",
                        help=("Specify data types to write for each sensor "
                              "type. Format: `-dtype sensor_0 int float`."))
    parser.add_argument("-seed", type=int, default=42,
                        help="Set the random seed for the generated data.")
    parser.add_argument("-runtime", type=float,
                        help=("Specify a maximum runtime before automatically "
                              "terminating the writer threads."))
    args = base.exhandler(examples, parser)
    if args is None:
        return

    #Verify that all the ports exist on the machine.
    from liveserial.monitor import enumerate_serial_ports
    from os import name
    allports = enumerate_serial_ports()
    failure = False
    for port in args["ports"]:
        if name != "nt":
            portname = "/dev/tty.{}".format(port)
        else: # pragma: no cover
            #We only have unit testing configured for unix-based systems at the
            #moment, so this check will never fire.
            portname = port
            
        if portname not in allports:
            msg.err("port {} does not exist locally.".format(port))
            failure = True
            
    if failure:
        return None
    
    sensors = {k: [] for k in args["ports"]}
    allsensors = []
    if args["sensors"]:
        current = args["ports"][0]
        for sensor in args["sensors"]:
            if sensor == "None":
                sensor = None
            allsensors.append(sensor)
            if sensor in args["ports"]:
                current = sensor
            else:
                sensors[current].append(sensor)
    args["sensors"] = sensors

    dtypes = {s: [] for s in allsensors}
    if args["dtype"]:
        current = None
        for dtype in args["dtype"]:
            if dtype == "None":
                dtype = None
            if dtype in allsensors:
                current = dtype
            else:
                dtypes[current].append(eval(dtype))
        for s in allsensors:
            if len(dtypes[s]) == 0:
                dtypes[s] = [int, float]
    args["dtype"] = dtypes
                
    return args

def _setup_simulator(args, port):
    """Sets up a simulator for the specified port.
    """
    from liveserial.simulator import ComSimulatorThread
    sensors = args["sensors"][port]
    dtypes = []
    for s in sensors:
        dtypes.append(tuple(args["dtype"][s]))

    simsig = ComSimulatorThread(port, sensors, dtypes, args["seed"])
    return simsig

def run(args, runtime=None):
    """Starts simulating the serial data in a separate thread.

    Args:
        runtime (float): how many seconds to run for before terminating.
    """
    if args is None: # pragma: no cover
        return

    simsigs = []
    for port in args["ports"]:
        simsigs.append(_setup_simulator(args, port))

    import signal
    def exit_handler(signal, frame): # pragma: no cover
        """Cleans up the serial communication, plotting and logging.
        """
        for simsig in simsigs:
            simsig.join(1)
        print("")
        exit(0)        
    signal.signal(signal.SIGINT, exit_handler)

    for simsig in simsigs:
        simsig.start()

    runtotal = 0.
    while all([sim.is_alive() for sim in simsigs]):
        #Here is the tricky bit; we want to join the threads for a second to
        #keep the main thread busy while everything is going on. More than a
        #second causes delay when the user sends SIGINT to stop the whole
        #process. We instead join on every thread for fractions of a second.
        for simsig in simsigs:
            simsig.join(1./len(simsigs), terminate=False)
        runtotal += 1.
        if runtime is not None and runtotal >= runtime:
            for simsig in simsigs:
                simsig.join(1)            

if __name__ == '__main__': # pragma: no cover
    #We skip the unit tests for this section because it is short and clear
    #and just calls methods that are being tested elsewhere.
    args = _parser_options()
    run(args, args["runtime"])



"""Class for monitoring a COM port using pyserial on a separate thread so that
that the main UI thread is not blocked or slowed. Code adapted from
https://github.com/mba7/SerialPort-RealTime-Data-Plotter/blob/master/com_monitor.py
"""
import threading, serial
try:
    from Queue import Empty
except ImportError: # pragma: no cover
    #Why couldn't they have called it queue from the start in py2?
    from queue import Empty
    
from liveserial import msg
class FormatInferrer(object):
    """Class that can infer the data types of an unknown sensor stream, provided
    they are consistent between calls.

    Args:
        infer_limit (int): number of raw lines to consider before reaching consensus
            on the format of the data (when inferring).
    """
    def __init__(self, infer_limit=15):
        self.infer_limit = infer_limit
        self._infer_count = 0
        """int: number of lines that have been analyzed alread for inferring data
        structure when no configuration is provided.
        """
        self._infer_keys = {}
        """dict: keys are sensor names, values are the indices at which the
        sensor key can be found in the raw line.
        """
        self.inferred = {}
        """dict: keys are inferred sensor ids, values are lists of casting types as for
        the implementation in :class:`liveserial.config.Sensor`. This dict only
        gets populated when no configuration is available and data structure has
        to be inferred from the incoming stream.

        """        

    def _infer_structure(self, raw):
        """Infers the structure of the raw data; returns the parsed line.
        
        Args:
            raw (list): of `str` values read from the line.
        """
        #We still try to infer some structure for the data by looking at
        #each value and trying to parse it as either int or float.
        dtype = []
        key = None
        for i, val in enumerate(raw):
            try:
                ival = int(val)
                dtype.append(int)
            except ValueError:
                try:
                    fval = float(val)
                    dtype.append(float)
                except ValueError:
                    key = i

        #Now that we have the general format, see if we can pick out the
        #sensor key from the string-valued entry.
        sensor = None
        if key is not None:
            sensor = raw[key]
        self._infer_keys[sensor] = key
                                         
        if sensor not in self.inferred:
            self.inferred[sensor] = dtype
        else:
            current = self.inferred[sensor]
            if current != dtype: # pragma: no cover
                #It turned out with unit testing that the consensus was reached
                #immediately so that this never fired.
                
                #We hope that the inferrence will reach a consensus at some
                #point. If it doesn't, do we take the latest point or stick
                #with the original? Probably, the first points will have the
                #highest likelihood to be incomplete.
                self.inferred[sensor] = dtype

    def parse(self, raw):
        """Attempts to parse the specified string values into one of the formats
        that has been inferred by this instance.

        Args:
            raw (list): of `str` values read from the line.
        """
        if self._infer_count < self.infer_limit:
            self._infer_structure(raw)
            self._infer_count += 1
            #We need to throw away the first points while we are inferring the
            #structure of the data.
            return (None, None)

        try:
            for k, v in self._infer_keys.items():
                if v is None:
                    continue
                if raw[v] == k:
                    result = []
                    for i in range(len(raw)):
                        if i != v:
                            caster = self.inferred[k][len(result)]
                            result.append(caster(raw[i]))

                    #This step makes sure that we have the right number of
                    #points to return.
                    if len(result) == len(self.inferred[k]):
                        return (result, k)
                    else: # pragma: no cover
                        # We have multiple checks to make sure the lengths are
                        # consistent, so this never fires in unit tests.
                        return (None, k)
            else:
                return ([t(r) for t, r in zip(self.inferred[None], raw)], None)
        except ValueError: # pragma: no cover
            #If the inference works correctly, we have already worked through
            #the formatting of the stream so that it should always parse
            #properly.
            return None, None

rxdelims = {}
"""dict: keys are port names; values are compiled regex objects. We wanted to
put these in the thread object, but then it can't be pickled, which makes issues
for the multiprocessing. Instead, we just have to do a dict lookup in the local
:meth:`Thread.run()`.
"""        
class ComMonitorThread(threading.Thread):
    """ A thread for monitoring a COM port. The COM port is 
        opened when the thread is started.
    
    Args:
        data_q (multiprocessing.Queue):
            Queue for received data. Items in the queue are
            (data, timestamp) pairs, where data is a binary 
            string representing the received data, and timestamp
            is the time elapsed from the thread's start (in 
            seconds).
        
        error_q (multiprocessing.Queue):
            Queue for error messages. In particular, if the 
            serial port fails to open for some reason, an error
            is placed into this queue.
        
        port (str):
            The COM port to open. Must be recognized by the 
            system.
        
        port_baud (int):
            Rate at which information is transferred in a communication channel
            (in bits/second).    

        stopbits (float): Serial communication parameter; one of (1, 1.5, 2).
        parity: (str): Serial communication parameter; one of ['N', 'E', 'O', 'M', 'S'].
        
        port_timeout (float):
            The timeout used for reading the COM port. If this
            value is low, the thread will return data in finer
            grained chunks, with more accurate timestamps, but
            it will also consume more CPU.
        listener (bool): specifies that this COMThread is a listener, which prints
            out the raw stream in real-time, but doesn't analyze it.
        infer_limit (int): number of raw lines to consider before reaching consensus
            on the format of the data (when inferring). If None, then a format
            inferrer is not created.
        virtual (bool): when True, additional serial port parameters are set so that
            the monitor can work with `socat` or other virtual ports.
        encoding (str): encoding type used to decode the byte stream from the serial
            port.
        delimiter (str): regex describing the sequence of characters used to
            separate columns of values in a single line from the serial port.
    Attributes:
        alive (threading.Event): event for asynchronously handling the reads from
          the serial port.
        serial_arg (dict): arguments used to contstruct the :class:`serial.Serial`.
        serial_port (serial.Serial): serial instance for communication.
        sensors (dict): keys are sensor names; values are
          :class:`liveserial.config.Sensor` instances used to
          parse raw lines read from the serial port.
        inferrer (FormatInferrer): for inferring the format in the absence of
            configured sensor structure.
    """
    def __init__(self, data_q, error_q, port, port_baud,
                 port_stopbits=serial.STOPBITS_ONE,
                 port_parity=serial.PARITY_NONE, port_timeout  = 0.01,
                 listener=False, virtual=False, infer_limit=15,
                 encoding="UTF-8", delimiter=r"\s"):
        threading.Thread.__init__(self)

        self.port = port
        self.serial_port = None
        self.serial_arg  = dict(port      = port,
                                baudrate  = int(port_baud),
                                stopbits  = float(port_stopbits),
                                parity    = port_parity,
                                timeout   = float(port_timeout))
        if virtual:
            msg.std("Running in virtual serial port mode.", 2)
            self.serial_arg["dsrdtr"] = True
            self.serial_arg["rtscts"] = True
            
        self.data_q   = data_q
        self.error_q  = error_q
        self.listener = listener
        self.encoding = encoding
        
        import re
        self.delimiter = delimiter
        global rxdelims
        rxdelims[port] = re.compile(delimiter)
        
        self.sensors = {}
        self._manual_sensors = False
        """bool: when True, the sensors list was constructed manually using a
        configuration file; otherwise, it was inferred from the first few lines
        of data from the serial port.
        """
        if infer_limit is not None:
            self.inferrer = FormatInferrer(infer_limit)
        else: # pragma: no cover
            self.inferrer = None
        self.alive    = threading.Event()
        self.alive.set()

    @staticmethod
    def from_port(port, port_baud=9600, virtual=False):
        """Returns a COMMonitor instance for the specified port using the
        default configurati of port parameters and with inferrence for the structure
        of the data.

        Args:
            port (str):
                The COM port to open. Must be recognized by the 
                system.
            
            port_baud (int):
                Rate at which information is transferred in a communication channel
                (in bits/second).    
        """
        from multiprocessing import Queue
        dataq = Queue()
        errorq = Queue()
        return ComMonitorThread(dataq, errorq, port, port_baud, virtual=virtual)
        
    @staticmethod
    def from_config(config, port, dataq=None, errorq=None, listener=False,
                    sfilter=None):
        """Returns a COMMonitor instance from the specified configuration
        parser.

        Args:
            config (ConfigParser): instance from which to extract the sensor list
              and port information. `str` is also allowed, in which case it
              should be the path to the config file to load.
            port (str): name of the port to configure for.
            data_q (multiprocessing.Queue):
                Queue for received data. Items in the queue are
                (data, timestamp) pairs, where data is a binary 
                string representing the received data, and timestamp
                is the time elapsed from the thread's start (in 
                seconds).            
            error_q (multiprocessing.Queue):
                Queue for error messages. In particular, if the 
                serial port fails to open for some reason, an error
                is placed into this queue.
            listener (bool): specifies that this COMThread is a listener, which prints
                out the raw stream in real-time, but doesn't analyze it.
            sfilter (list): of sensor names that should be *included* in the
              monitor. By default, all sensors in the config are included that
              match the port.
        Returns:
            ComMonitorThread: instance created using the configuration
            parameters.
        """        
        from multiprocessing import Queue
        #We allow the user to choose to use a common data and error queue
        #between all the threads. If they don't specify one, then we will just
        #create a new one.

        #This method is usually called from the entry script, which allows all
        #the ports to share the same multiprocessing queues by default.
        if dataq is None: # pragma: no cover
            dataq = Queue()
        if errorq is None: # pragma: no cover
            errorq = Queue()

        from liveserial.config import ports
        params = ports(config, port)
        vtext = "{}: using {} as config-set serial parameters."
        msg.std(vtext.format(port, params))
        result = ComMonitorThread(dataq, errorq, port, listener=listener,
                                  **params)

        from liveserial.config import sensors, Sensor
        sdict = sensors(config, port=port, monitor=result)
        for sensor, instance in sdict.items():
            if (sfilter is None or sfilter == "all") or sensor in sfilter:
                result.add_sensor(sensor, instance)

        return result

    def add_sensor(self, name, sensor):
        """Adds a sensor to the COM monitors active list. Sensors in the active
        list have their data parsed and pushed to the data queue. If no sensors
        are added to the list, the monitor will try to infer the sensor
        configuration using the first few raw data samples.

        Args:
            sensor (liveserial.config.Sensor): instance with parsed
              configuration values.
        """
        self._manual_sensors = True
        self.sensors[name] = sensor
        
    def run(self):
        """Starts the COM monitoring thread. If an existing serial connection is
        open, it will be closed and a new one will be created. The monitoring
        will continue indefinitely until :meth:`join` is called.
        """
        try:
            if self.serial_port: # pragma: no cover
                self.serial_port.close()
            self.serial_port = serial.Serial(**self.serial_arg)
            msg.info("Serial port communication enabled.", 2)
        except serial.SerialException as e: # pragma: no cover
            self.error_q.put(e.strerror)
            return

        from time import time
        start = time()
        lastlisten = None
        decoderr = 0 #Number of types the decode has failed.
        rxsplit = rxdelims[self.port]
        
        while self.alive.isSet():
            line = self.serial_port.readline()                
            if self.listener:
                if lastlisten is None or time() - lastlisten > 0.05:
                    print(line)
                    lastlisten = time()
                #This is crazy to me... The if statement above and its contents
                #are covered by the unit tests, but coverage claims that this
                #continue statement is not...
                continue # pragma: no cover

            try:
                raw = line.decode(self.encoding).strip()
            except: # pragma: no cover
                #It is too much hassle to get our port data simulator to simulate garbage
                #for now. TODO: update the simulator to randomly spew out garbage.
                if len(line) > 0:
                    decoderr += 1
                    if decoderr < 3:
                        emsg = "Couldn't decode line {} using {}."
                        msg.warn(emsg.format(line, self.encoding), -1)

            raw = rxsplit.split(raw)
            if len(raw) == 0: # pragma: no cover
                #No data read from the stream.
                continue
            
            vals, sensor = None, None
            if self._manual_sensors:
                for osensor in self.sensors.values():
                    vals, skey = osensor.parse(raw), osensor.key
                    sensor = osensor.name
                    #The parsing will return None if the raw line does not apply
                    #to its configuration. The moment we found the one that does
                    #apply, there is not sense in still checking.
                    if vals is not None:
                        break
            else:
                #We try to infer the structure of the data from the raw line.
                if self.inferrer is not None:
                    vals, sensor = self.inferrer.parse(raw)

            if vals is not None and len(vals) > 0:
                if sensor is None:
                    #We use the id of the com thread (which corresponds
                    #one-to-one with the serial ports) as the sensor key; that
                    #way, multiple inferred, null-key sensors from different
                    #ports can still be differentiated in the logs.
                    sensor = id(self)
                timestamp = time() - start
                self.data_q.put((sensor, timestamp) + tuple(vals))
            #else:
            # There must have been a communication glitch; just ignore
            # this data point.
            
        # clean up
        if self.serial_port:
            self.serial_port.close()
            
    def join(self, timeout=None, terminate=True):
        """Tells the thread monitoring the COM port to clean up and return.

        Args:
            timeout (float): number of seconds (or fractions of seconds) to wait
              until returning. If `None`, then the operation will
              block until the thread terminates. See also
              :meth:`threading.Thread.join`.
            terminate (bool): when True, the data collection is told to stop before
              trying to join the underlying thread; otherwise, the thread will
              keep processing data until join is called with terminate=True.
        """
        if terminate:
            self.alive.clear()
        threading.Thread.join(self, timeout)

def enumerate_serial_ports():
    """Scans for available serial ports.

    Returns:         
    list: of `str` with  the availables port names.
    """
    from os import name
    from glob import glob
    if name  == 'nt': # pragma: no cover
        #We don't have CI setup for windows at the moment.
        from os import waitpid, path
        from subprocess import Popen, PIPE
        cmd = 'powershell -c "[System.IO.Ports.SerialPort]::getportnames()"'
        pps = Popen(cmd, stdout=PIPE, stderr=PIPE)
        output, error = pps.communicate()

        if len(error) > 0:
            msg.error(''.join([s.decode("UTF-8") for s in error]))

        result = []
        for oline in output.decode("UTF-8").strip().split():
            if len(oline) > 0:
                result.append(oline)
        return result
    else:
        return glob('/dev/tty.*')
    
def get_all_from_queue(Q):
    """Generator to yield one after the others all items currently in the queue Q,
        without any waiting.

    Args:
        Q (Queue.Queue): queue to empty items from.
    """
    msg.info("Retrieving entire queue.", 3)
    try:
        while True:
            yield Q.get_nowait()
    except Empty:
        raise StopIteration

def get_item_from_queue(Q, timeout=0.01):
    """ Attempts to retrieve an item from the queue Q. If Q is
        empty, None is returned.

    Args:
        Q (Queue.Queue): queue to get an item from.
        timeout (float):
            Blocks for 'timeout' seconds in case the queue is empty,
            so don't use this method for speedy retrieval of multiple
            items (use get_all_from_queue for that).
    """
    try: 
        item = Q.get(True, 0.01)
    except Empty: 
        return None
    return item

class LiveDataFeed(object):
    """A simple "live data feed" abstraction that allows a reader to read the most
    recent data and find out whether it was updated since the last read.

    Attributes:
        has_new_data (dict): A boolean attribute telling the reader whether the
          data was updated since the last read; keyed by sensor identifier.
        cur_data (dict): most recent data point placed on the feed; keyed by the
          sensor identifier.
    """
    def __init__(self):
        self.cur_data = {}
        self.has_new_data = {}
        
    def add_data(self, sensor, data):
        """Add new data to the feed.
        
        Args:
        sensor (str): sensor identifier for the data point.
        """
        self.cur_data[sensor] = data
        self.has_new_data[sensor] = True
    
    def read_data(self, sensor):
        """Returns the most recent data."""
        self.has_new_data[sensor] = False
        return self.cur_data[sensor]



"""Thread for simulating writes to a virtual port.
"""
import threading
from liveserial import msg
class ComSimulatorThread(threading.Thread):
    """Simulates a sine wave, masquerading as a separate COM port on the machine so
    that we can unit test the code against it.

    Args:
        port (str): name of the simulated port to write to.
        dataform (dict): keys are sensor ids; values are tuples of `type` that
          specifies how a row of simulated data will look when written to the COM
          port.
        sensors (list): of `str` giving sensor ids for which data will be randomly
          generated with equal probability between each sensor. If the sensor id
          is `None`, then no sensor key will be written to the stream.
        seed (int): random seed so the values are predictable.

    Attributes:
        alive (threading.Event): event for asynchronously handling the reads from
          the serial port.

    Examples:
        Write three random columns with data types `int`, `float` and `float` to
        `COM3` *without* any sensor identifying column. 

        >>> from liveserial.simulator import ComSimulatorThread
        >>> cst = ComSimulatorThread("COM3", sensors=[None], 
            ... dataform=[(int, float, float)])
        >>> cst.start()

        Note that the writing happens in its own thread (:class:`ComSimulatorThread`
        inherits from :class:`threading.Thread`), so it will run indefinitely if the
        thread is not joined. A typical use case is:

        >>> import signal
        >>> def exit_handler(signal, frame):
            ... cst.join(1)
        >>> signal.signal(signal.SIGINT, exit_handler)
        >>> cst.start()

        This wires the signal interrupt request (usually raised by pressing ^C) to
        join the simulator thread.
    """
    def __init__(self, port="lscom-w", sensors=["W", None, "K"],
                 dataform=[(int, float), (float, float, float), (int, float)],
                 seed=42):
        threading.Thread.__init__(self)
        self.dataform = {s: d for s, d in zip(sensors, dataform)}
        self.sensors = sensors
        from os import name
        if name  == 'nt': # pragma: no cover
            self.port = port
        else:
            self.port = "/dev/tty.{}".format(port)

        from serial import Serial
        self.serial = Serial(self.port, 9600, dsrdtr=True, rtscts=True)
        self.seed = seed
        self.alive = threading.Event()
        self.alive.set()
        
    def run(self):
        """Starts simulating the communication. This method should not be called
        directly. Instead, it is invoked automatically when :meth:`start` is
        called on this thread object.
        """
        import random, time, math
        import os
        #Seed the random number generator so that it always produces the same
        #values for the random variables.
        random.seed(self.seed)
        
        while self.alive.isSet():
            #Choose one of the sensors at random to generate data for.
            randsense = int(len(self.sensors)*random.random())
            sensor = self.sensors[randsense]
            if sensor is not None:
                raw = [sensor]
            else:
                raw = []
                
            for t in self.dataform[sensor]:
                if t is int:
                    raw.append(random.randint(0, 100))
                elif t is float:
                    raw.append(random.randint(-1, 1) + random.random())

            data = ' '.join([str(d) for d in raw]) + os.linesep
            #Usually people encode with UTF-8. However, we know that our data is
            #really simple and ASCII takes less space.
            x = self.serial.write(data.encode("ascii"))
            time.sleep(0.0025)

        if self.serial:
            self.serial.close()

    def join(self, timeout=None, terminate=True):
        """Tells the thread simulating the COM port to clean up and return.

        Args:
            timeout (float): number of seconds (or fractions of seconds) to wait
              until returning. If `None`, then the operation will
              block until the thread terminates. See also
              :meth:`threading.Thread.join`.
        """
        if terminate:
            self.alive.clear()
            self.serial.cancel_write()
            self.serial.flushOutput()
        threading.Thread.join(self, timeout)



#!/usr/bin/env python
try:
    from setuptools import setup
    args = {}
except ImportError:
    from distutils.core import setup
    print("""\
*** WARNING: setuptools is not found.  Using distutils...
""")

from setuptools import setup
try:
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rst')
except ImportError:
    print("warning: pypandoc module not found, could not convert Markdown to RST")
    read_md = lambda f: open(f, 'r').read()

from os import path
setup(name='live-serial',
      version='0.1.4',
      description='Real-time serial port plotter/logger.',
      long_description= "" if not path.isfile("README.md") else read_md('README.md'),
      author='Conrad W Rosenbrock',
      author_email='rosenbrockc@gmail.com',
      url='https://github.com/rosenbrockc/live-serial',
      license='MIT',
      setup_requires=['pytest-runner',],
      tests_require=['pytest', 'python-coveralls'],
      install_requires=[
          "argparse",
          "termcolor",
          "numpy",
          "matplotlib",
          "pyserial",
      ],
      packages=['liveserial'],
      scripts=['liveserial/livemon.py'],
      package_data={'liveserial': []},
      include_package_data=True,
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Intended Audience :: Science/Research',
          'Natural Language :: English',
          'Operating System :: MacOS',
          'Operating System :: Microsoft :: Windows',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
      ],
     )



#!/usr/bin/python
from liveserial import msg
def examples():
    """Prints examples of using the script to the console using colored output.
    """
    script = "LIVE-SERIAL Real-time Serial Port Plotter/Logger"
    explain = ("For any device connected to a serial port (including USB), "
               "it is useful to see and log the data from the port in "
               "real-time (for example in response to human-interaction. "
               "This script sets up live monitoring of a serial port on a "
               "separate thread and plots the data as it is received. Logging "
               "can also be enabled in real-time.")
    contents = [(("Start plotting the data from the COM3 serial port."), 
                 "livemon.py COM3",
                 "If '-logdir' is not specified, *no* data will be logged."),
                (("Plot *and* log the data from /dev/tty."),
                 "livemon.py /dev/tty -logdir ~/sensordata", ""),
                ("Log the data, but don't generate a live plot.",
                 "livemon.py COM3 -noplot -logdir C:\\sensordata\\", ""),
                ("List the available serial ports on the system.",
                 "livemon.py list", ""),
                ("Use the config file `custom.cfg` to log and plot the data. "
                 "Plot any sensors listed in that file.",
                 "livemon.py auto -config custom.cfg",
                 "If your file is called `sensors.cfg`, you can just use "
                 "`livemon.py auto`.")]
    required = ("REQUIRED: working serial port.")
    output = ("RETURNS: plot window; for logging-only mode, the data being "
              "logged is also periodically printed to stdout.")
    details = ("The plotting uses `matplotlib` with the default configured "
               "backend. If you want a different backend, set the rc config "
               "for `matplotlib` using online documentation. However, many "
               "backends don't play well with the animation (depending on OS "
               "type and version, etc., etc.; so use carefully.")
    outputfmt = ("")

    msg.example(script, explain, contents, required, output, outputfmt, details)

def _list_serial(ports=None):
    """Lists all of the available serial ports on the local machine. If `port`
    is specified, then returns True if port is in the list.

    Args:
    ports (list): name(s) of ports to check existence for.

    Returns:
    bool: specifying whether the given port is in the list, or
    list: of available ports on this machine.
    """
    from liveserial.monitor import enumerate_serial_ports
    available = enumerate_serial_ports()
    if ports is not None:
        result = True
        for port in ports:
            if port not in available:
                msg.err("Port '{}' is not valid.".format(port))
                result = False
        return result
    else:
        return available

def _list_config_ports(configfile):
    """Lists all the ports explicitly mentioned in the config file.
    """
    try:
        from configparser import ConfigParser
    except ImportError: # pragma: no cover
        #Renaming of modules to lower case in python 3.
        from ConfigParser import ConfigParser

    config = ConfigParser()
    config.readfp(open(configfile))
    
    from fnmatch import fnmatch
    result = []
    for section in config.sections():
        if fnmatch(section, "sensor.*"):
            if config.has_option(section, "port"):
                port = config.get(section, "port")
                if port not in result:
                    result.append(port)
        elif fnmatch(section, "port.*"):
            port = section[5:]
            if port not in result: # pragma: no cover
                result.append(port)

    return result

script_options = {
    "port": {"nargs": "+",
             "help": "Name of the port(s) to plot/log."},
    "-noplot": {"action": "store_true",
                "help": "Don't plot the data; only log it."},
    "-autoconf": {"action": "store_true",
                  "help": ("Runs in automatic mode by loading a "
                           "configuration file in the current directory "
                           "called 'sensors.cfg'.")},
    "-config": {"help": ("Specify a configuration file to get sensor "
                         "setup information from.")},
    "-logdir": {"help": ("Path to the directory where sensor data will "
                         "be logged.")},
    "-baudrate": {"type": int, "default": 9600,
                  "help": ("Rate at which information is transferred in a "
                           "communication channel (in bits/second).")},
    "-stopbits": {"type": float, "default": 1., "choices": [1, 1.5, 2],
                  "help": "Serial communication parameter."},
    "-parity": {"type": str, "default": 'N',
                "choices": ['N', 'E', 'O', 'M', 'S'],
                "help": "Serial communication parameter."},
    "-timeout": {"type": float, "default": 0.01,
                 "help": ("The timeout used for reading the COM port. If "
                          "this value is low, the thread will return data "
                          "in finer grained chunks, with more accurate "
                          "timestamps, but it will also consume more CPU.")},
    "-refresh": {"type": int, "default": 100,
                 "help": ("How often (in milliseconds) to plot new data "
                          "obtained from the serial port.")},
    "-buffertime": {"type": float, "default": 25,
                    "help": ("How often (in milliseconds) to query buffered data "
                             "obtained from the serial port.")},
    "-logfreq": {"type": float, "default": 10,
                 "help": ("How often (in *seconds*) to save the buffered "
                          "data points to CSV.")},
    "-method": dict(default="average", choices=["last", "average"],
                    help=("Specifies how buffered data is aggregated each "
                          "time it is read from the serial port.")),
    "-listen": dict(action="store_true",
                    help=("Prints the raw output from the serial port "
                          "instead of plotting and logging it. Useful "
                          "for debugging port connection issues.")),
    "-virtual": dict(action="store_true",
                     help=("Specifies that the port being connected to is "
                           "virtual (e.g., with `socat`), which changes the "
                           "parameters for connection.")),
    "-sensors": dict(nargs="+", default="all",
                     help="Filter the list of sensors being logged/plotted."),
    "-maxpts": dict(type=int, default=100,
                    help=("Maximum number of values to keep in the plot "
                          "for each sensor")),
    "-window": dict(type=float, default=20.,
                    help="Width of window in time units."),
    "-wait": dict(type=float, default=1.,
                  help=("Number of seconds to wait before failing because "
                        "no data is present on the stream for plotting."))
    }
"""dict: default command-line arguments and their
    :meth:`argparse.ArgumentParser.add_argument` keyword arguments.
"""

def _parser_options():
    """Parses the options and arguments from the command line."""
    #We have two options: get some of the details from the config file,
    import argparse
    from liveserial import base
    pdescr = "Real-time serial port plotter/logger."
    parser = argparse.ArgumentParser(parents=[base.bparser], description=pdescr)
    for arg, options in script_options.items():
        parser.add_argument(arg, **options)
        
    args = base.exhandler(examples, parser)
    if args is None:
        return

    #Handle the automatic configuration file setup.
    if args["autoconf"] or (args["port"]==["auto"] and not args["config"]):
        from os import path
        args["config"] = path.abspath("sensors.cfg")
        msg.info("Using {} as auto-selected config path".format(args["config"]))

    if args["config"]:
        #Try and override any of the other options with those of the config
        #file, if it exists.
        from os import path
        args["config"] = path.abspath(path.expanduser(args["config"]))
        from liveserial.config import script
        args.update(script(args["config"]))
        msg.info("Using script args from config {}.".format(args["config"]))
        
    if args["port"] == ["list"]:
        msg.okay("Available Serial Ports")
        for port in _list_serial():
            msg.info("  {}".format(port))
        return None
    elif args["port"] == ["auto"]:
        #We load all the ports present in the config file and assume that the
        #user meant all of those.
        args["port"] = _list_config_ports(args["config"])
    else:
        if not _list_serial(args["port"]):
            return None

    #Convert the units for the buffer and refresh times.
    args["refresh"] /= 1000.
    args["buffertime"] /= 1000.
    
    if args["noplot"] and not args["logdir"]: # pragma: no cover
        msg.warn("Data will only be logged if `-logdir` is specified.", -1)

    return args

def _get_com(args):
    """Gets a list of configured COM ports for serial communication.
    """
    from liveserial.monitor import ComMonitorThread as CMT
    from multiprocessing import Queue
    dataq, errorq = Queue(), Queue()
    result = []
    msg.info("Starting setup of ports {}.".format(args["port"]), 2)
    if args["config"]:
        for port in args["port"]:
            if port.lower() != "aggregate":
                #The aggregate port name is just a shortcut so that we can plot
                #transforms between multiple sensor streams. It doesn't actually
                #represent a physical port that will be monitored.
                com = CMT.from_config(args["config"], port, dataq, errorq,
                                      args["listen"], args["sensors"])
                result.append(com)                               
    else:
        for port in args["port"]:
            com = CMT(dataq, errorq, port, args["baudrate"],
                      args["stopbits"], args["parity"], args["timeout"],
                      args["listen"], args["virtual"])
            result.append(com)
    return result

def _com_start(coms):
    """Starts the serial port communication thread using the specified object.

    Args:
        coms (list): of :class:`monitor.ComMonitorThread` instances to start
          running.
    """
    from liveserial.monitor import get_item_from_queue
    for i, com in enumerate(coms):
        com.start()
        msg.okay("COM monitoring thread {} started.".format(i), 2)

        #Even though the thread is going, it doesn't mean that everything is working
        #the way we hope. Check the first item. We have to sleep to give it time to
        #initialize.
        com.join(0.05, terminate=False)
        com_error = get_item_from_queue(com.error_q)
        if com_error is not None: # pragma: no cover
            #We can't easily simulate failure on the serial port to test this message.
            msg.err("monitor thread error--{}".format(com_error))
            coms[i] = None
    
def run(args, runtime=None, testmode=False):
    """Starts monitoring of the serial data in a separate thread. Starts the
    plotting or logging based on command-line args.

    Args:
        args (dict): result of :func:`_parser_options` call that parses the
          command-line arguments. Could also be :data:`script_options` to use all
          default values.
        runtime (float): how many seconds to run for before terminating.
        testmode (bool): when True, the plotting backend is changed so that no
          window is produced; plotting functions are altered slightly to make
          testing possible.
    """
    from liveserial.base import set_testmode
    set_testmode(testmode)    
    #When args is None, it means that examples or help or equivalent wants to
    #cancel the execution of the script.
    if args is None:
        return
    
    #Get the serial port for communications; this also tests that we are getting
    #data.
    vardir = {}
    coms = _get_com(args)

    #The data feed keeps track of the latest, aggregated data selected from the
    #buffer in the com thread.
    from liveserial.monitor import LiveDataFeed
    feed = LiveDataFeed()
    
    #Logging queries the com thread buffer to get data, and then aggregates it
    #and puts it on the live data feed. Optionally, the data is also
    #periodically saved to CSV.
    from liveserial.log import Logger
    dataqs = [c.data_q for c in coms]
    #The logger prints values to screen if it isn't running in plotting
    #mode. Plotting mode means that the plot window is present, or that
    #'-listen' was specified.
    plotting = args["listen"] or (not args["noplot"])
    logger = Logger(args["buffertime"], dataqs, feed,
                    args["method"], args["logdir"], args["logfreq"],
                    plotting, args["config"],
                    aggregate="aggregate" in args["port"])
    
    import signal
    def exit_handler(signal, frame): # pragma: no cover
        """Cleans up the serial communication and logging.
        """
        msg.warn("SIGINT >> cleaning up threads.", -1)
        logger.stop()
        for com in coms:
            com.join(1)
        if plotter is not None:
            plotter.stop()
        print("")
        exit(0)
        #Matplotlib's cleanup code for animations is lousy--it doesn't
        #work. I tried calling the private _stop() in a relevant scope and
        #it still let the application hang.
    signal.signal(signal.SIGINT, exit_handler)

    #Add the local variables to the dictionary. This dict is passed in by unit
    #tests usually, which want to investigate the values in each object.
    if vardir is not None:
        vardir["feed"] = feed
        vardir["com"] = coms
        vardir["logger"] = logger
        
    #Now that we actually have a way to quit the infinite loop, we can start the
    #data acquisition process.
    plotter = None
    _com_start(coms)
    if any([com is None for com in coms]): # pragma: no cover
        msg.err("One of the COM threads didn't initialize properly.")
        msg.info("COM Threads that were Okay")
        for com in coms:
            if com is not None:
                msg.std(com.port)
        return

    #Wait until we have some data before the logger gets put to work with it.
    if not args["listen"]:
        from time import sleep
        tries = 0
        while (not all([not com.data_q.empty() for com in coms])
               and tries < 10): # pragma: no cover
            #If we don't need this delay, it shouldn't trigger unit test
            #problems.
            sleep(0.05)
            tries += 1
        logger.start()

    #The plotter looks at the live feed data to plot the latest aggregated
    #points.
    if not args["noplot"] and not args["listen"]:
        tries = 0
        maxtries = args["wait"]/0.05
        while not logger.ready(0.05) and tries <= maxtries:
            tries += 1

        from liveserial.plotting import Plotter
        import matplotlib.pyplot as plt
        plotter = Plotter(feed, args["refresh"], args["maxpts"],
                          args["window"], testmode, logger)
        if vardir is not None:
            vardir["plotter"] = plotter
        if not testmode: # pragma: no cover
            plt.show()

    _runtime = 0
    while all([com.is_alive() for com in coms]):
        #Here is the tricky bit; we want to join the threads for a second to
        #keep the main thread busy while everything is going on. More than a
        #second causes delay when the user sends SIGINT to stop the whole
        #process. We instead join on every thread for fractions of a second.
        for com in coms:
            com.join(1./len(coms), terminate=False)
        _runtime += 1.
        if runtime is not None:
            if _runtime >= runtime:
                logger.stop()
                for com in coms:
                    com.join(1)
                if plotter is not None:
                    plotter.stop()

    return vardir

if __name__ == '__main__': # pragma: no cover
    run(_parser_options())
